D_COMMAND = """
d() {{
    declare _dtags_usage="{usage}"
    declare _dtags_version="{version}"
    declare _dtags_description="{description}"
    if [[ -t 1 ]]
    then
        declare _dtags_goto_msg="{goto_msg_tty}"
        declare _dtags_arg_err="{arg_err_tty}"
        declare _dtags_dest_err="{dest_err_tty}"
        declare _dtags_input_err="{input_err_tty}"
        declare _dtags_idx_err="{index_err_tty}"
        declare _dtags_prompt="{prompt_tty}"
        declare _dtags_choice="{choice_tty}"
    else
        declare _dtags_goto_msg="{goto_msg}"
        declare _dtags_arg_err="{arg_err}"
        declare _dtags_dest_err="{dest_err}"
        declare _dtags_input_err="{input_err}"
        declare _dtags_idx_err="{index_err}"
        declare _dtags_prompt="{prompt}"
        declare _dtags_choice="{choice}"
    fi
    if [[ $# -lt 1 ]]
    then
        printf "$_dtags_goto_msg" "$HOME"
        cd $HOME
        return 0
    elif [[ $1 = --help ]]
    then
        printf "$_dtags_usage$_dtags_description"
        return 0
    elif [[ $1 = --version ]]
    then
        printf "Version $_dtags_version\n"
        return 0
    elif [[ $1 = -* ]]
    then
        printf "$_dtags_arg_err" "$_dtags_usage" "$1"
        return 2
    elif [[ $# -gt 1 ]]
    then
        printf "%sToo many arguments\n" "$_dtags_usage"
        return 2
    fi
    declare -a _dtags_dirs
    PREV_IFS=$IFS
    IFS=$'\n'
    _dtags_dirs=($(grep -F ,"$1", {mapping_file} | cut -d',' -f1))
    IFS=$PREV_IFS
    declare _dtags_count=${{#_dtags_dirs[@]}}
    if [[ $_dtags_count -eq 0 ]]
    then
        if [[ -d $1 ]]
        then
            declare _dtags_dir="$1"
        else
            printf "$_dtags_dest_err" "$_dtags_usage" "$1"
            return 2
        fi
    elif [[ $_dtags_count = 1 ]]
    then
        declare _dtags_dir=${{_dtags_dirs[0]}}
    else
        for i in `seq $_dtags_count`
        do
            printf "$_dtags_choice" "$i" "${{_dtags_dirs[$i-1]}}"
        done
        printf "$_dtags_prompt" "$_dtags_count"
        declare _dtags_idx=1
        read _dtags_idx
        if [[ ! $_dtags_idx =~ ^-?[0-9]+$ ]]
        then
            printf "$_dtags_input_err" "$_dtags_idx"
            return 2
        elif [[ $_dtags_idx -gt $_dtags_count ]] || [[ $_dtags_idx -lt 1 ]]
        then
            printf "$_dtags_idx_err" "$_dtags_idx"
            return 2
        fi
        declare _dtags_dir=${{_dtags_dirs[$_dtags_idx-1]}}
    fi
    if [[ -d $_dtags_dir ]]
    then
        printf "$_dtags_goto_msg" "$_dtags_dir"
        cd "$_dtags_dir"
        return 0
    else
        printf "$_dtags_dest_err" "$_dtags_usage" "$_dtags_dir"
        return 2
    fi
}}


__dtags_d_command() {{
    declare cur=${{COMP_WORDS[COMP_CWORD]}}
    if [[ ${{COMP_CWORD}} -eq 1 ]]
    then
        if [[ -f "{tags_file}" ]]
        then
            COMPREPLY+=($(compgen -W "`cat {tags_file}`" -- $cur))
        fi
    fi
}}


__dtags_e_command() {{
    declare cur=${{COMP_WORDS[COMP_CWORD]}}
    if [[ ${{COMP_CWORD}} -eq 1 ]]
    then
        if [[ -f "{tags_file}" ]]
        then
            COMPREPLY+=($(compgen -W "`cat {tags_file}`" -- $cur))
        fi
        COMPREPLY+=($(compgen -W "-p" -- $cur))
    elif [[ ${{COMP_CWORD}} -eq 2 && ${{COMP_WORDS[1]}} = -p ]]
    then
        if [[ -f "{tags_file}" ]]
        then
            COMPREPLY+=($(compgen -W "`cat {tags_file}`" -- $cur))
        fi
    fi
}}


__dtags_tag_untag_commands() {{
    declare cur=${{COMP_WORDS[COMP_CWORD]}}
    if [[ ${{COMP_CWORD}} -eq 1 ]]
    then
        _filedir -d
    elif [[ -f "{tags_file}" ]] && [[ ! ${{COMP_WORDS[1]}} = -* ]]
    then
        COMPREPLY+=($(compgen -W "`cat {tags_file}`" -- $cur))
    fi
}}


__dtags_main_command() {{
    declare cur=${{COMP_WORDS[COMP_CWORD]}}
    if [[ ${{COMP_CWORD}} -eq 1 ]]
    then
        COMPREPLY+=($(compgen -W "list reverse shell edit clean" -- $cur))
    elif [[ ${{COMP_WORDS[1]}} = list ]] || [[ ${{COMP_WORDS[1]}} = reverse ]]
    then
        if [[ -f "{tags_file}" ]]
        then
            COMPREPLY+=($(compgen -W "`cat {tags_file}`" -- $cur))
        fi
        _filedir -d
    fi
}}

complete -o dirnames -F __dtags_e_command e
complete -o dirnames -F __dtags_d_command d
complete -o dirnames -F __dtags_main_command dtags
complete -o dirnames -F __dtags_tag_untag_commands tag untag
"""



from __future__ import unicode_literals, print_function

import os
import sys
import atexit
import signal

from dtags import style
from dtags.chars import get_invalid_path_chars, get_invalid_tag_chars
from dtags.config import load_mapping, save_mapping
from dtags.utils import close_stdio, abort, finish, expand
from dtags.version import VERSION

USAGE = """Usage:
  tag <dir> [<tag>...]
  tag --help
  tag --version
"""
DESCRIPTION = """
Arguments:
  dir     The directory path
  tag     The directory tag

Tag the target directory. If no tag names are given,
the basename of the directory path is used instead."""


def main():
    atexit.register(close_stdio)
    signal.signal(signal.SIGPIPE, signal.SIG_DFL)
    args = sys.argv[1:]
    if not args:
        finish(USAGE + DESCRIPTION)
    head, tail = args[0], args[1:]
    if head == '--help':
        finish(USAGE + DESCRIPTION)
    elif head == '--version':
        finish('Version ' + VERSION)
    elif head.startswith('-'):
        abort(USAGE + 'Invalid argument: ' + style.bad(head))
    path = expand(head)
    invalid_path_chars = get_invalid_path_chars(path)
    if invalid_path_chars:
        abort('Directory path {} contains bad characters {}'.format(
            style.bad(path), style.bad_chars(invalid_path_chars)
        ))
    if not os.path.isdir(path):
        abort('Invalid directory ' + style.bad(path))
    mapping, excluded = load_mapping()
    tags_added = set()
    if not tail:
        tail.append(os.path.basename(path))
    for tag in tail:
        if not tag[0].isalpha():
            abort('Tag name {} does not start with an alphabet'
                  .format(style.bad(tag)))
        if ' ' in tag:
            abort('Tag name {} contains whitespaces'.format(style.bad(tag)))
        invalid_tag_chars = get_invalid_tag_chars(tag)
        if invalid_tag_chars:
            abort('Tag name {} contains bad characters {}'.format(
                style.bad(tag), style.bad_chars(invalid_tag_chars)
            ))
        if path not in mapping[tag]:
            mapping[tag].add(path)
            tags_added.add(tag)
    if tags_added or excluded:
        save_mapping(mapping)
    if excluded:
        print('Cleaned the following invalid entries:\n' + excluded + '\n')
    if not tags_added:
        finish('Nothing to do')
    else:
        finish(style.path(path) + ' ' + ' '.join(
            style.sign('+') + style.tag(tag) for tag in sorted(tags_added)
        ))



from __future__ import unicode_literals, print_function

import sys
import atexit
import signal

from dtags import style
from dtags.config import load_mapping, save_mapping
from dtags.utils import close_stdio, finish, expand, abort
from dtags.version import VERSION

USAGE = """Usage:
  untag <dir> [<tag>...]
  untag --help
  untag --version
"""
DESCRIPTION = """
Arguments:
  dir     The directory path
  tag     The directory tag

Remove one or more tags from the target directory.
If no tag names are given, ALL tags mapped to the
target directory are removed."""


def main():
    atexit.register(close_stdio)
    signal.signal(signal.SIGPIPE, signal.SIG_DFL)
    mapping, excluded = load_mapping()
    args = sys.argv[1:]
    if not args:
        finish(USAGE + DESCRIPTION)
    head, tail = args[0], args[1:]
    if head == '--help':
        finish(USAGE + DESCRIPTION)
    elif head == '--version':
        finish('Version ' + VERSION)
    elif head.startswith('-'):
        abort(USAGE + 'Invalid argument: ' + style.bad(head))
    path = expand(head)
    tags_removed = set()
    for tag in tail if tail else mapping.keys():
        if path in mapping[tag]:
            mapping[tag].remove(path)
            tags_removed.add(tag)
    if not tags_removed:
        finish('Nothing to do')
    save_mapping(mapping)
    if excluded:
        print('Cleaned the following invalid entries:\n' + excluded + '\n')
    finish(style.path(path) + ' ' + ' '.join(
        style.sign('-') + style.tag(tag) for tag in sorted(tags_removed)
    ))



from __future__ import unicode_literals

from dtags import style

USAGE = """Usage:
  d [<dir>|<tag>]
  d --help
  d --version
"""
DESCRIPTION = """
Arguments:
  dir         The directory path
  tag         The directory tag

Options:
  --help      Display the help menu
  --version   Display the version

Change directory. If the destination is not
specified it defaults to the home directory.
If multiple directories are associated with
the same tag, a selection menu is displayed.
"""

ARG_ERR = '%sInvalid argument: %s\n'
CHOICE = '%s: %s\n'
DEST_ERR = '%sInvalid destination: %s\n'
GOTO_MSG = 'Going to %s\n'
INDEX_ERR = 'Index out of range: %s\n'
INPUT_ERR = 'Invalid input: %s\n'
PROMPT = '\nSelect directory (1 - %s): '

# Style the messages if the output is going to a terminal
tty = True
ARG_ERR_TTY = '%sInvalid argument: ' + style.bad('%s\n', tty=tty)
CHOICE_TTY = style.msg('%s: ', tty=tty) + style.path('%s\n', tty=tty)
DEST_ERR_TTY = '%sInvalid destination: ' + style.bad('%s\n', tty=tty)
GOTO_MSG_TTY = style.msg('Going to ', tty=tty) + style.path('%s\n', tty=tty)
INDEX_ERR_TTY = 'Index out of range: ' + style.bad('%s\n', tty=tty)
INPUT_ERR_TTY = 'Invalid input: ' + style.bad('%s\n', tty=tty)
PROMPT_TTY = style.msg('\nSelect directory (1 - %s): ', tty=tty)



from __future__ import absolute_import, print_function, unicode_literals

# Allowed characters in tag names
TAG_CHARS = {
    'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o',
    'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', 'A', 'B', 'C', 'D',
    'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S',
    'T', 'U', 'V', 'W', 'X', 'Y', 'Z', '0', '1', '2', '3', '4', '5', '6', '7',
    '8', '9', '-', '_', '.'
}
# Allowed characters in directory paths
PATH_CHARS = {
    'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o',
    'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', 'A', 'B', 'C', 'D',
    'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S',
    'T', 'U', 'V', 'W', 'X', 'Y', 'Z', '0', '1', '2', '3', '4', '5', '6', '7',
    '8', '9', '-', '_', '.', '/', '~', ' '
}


def get_invalid_tag_chars(tag):
    """Return any invalid characters in the directory tag.

    :param tag: the directory tag
    :returns: set of all invalid characters
    """
    return set(char for char in tag if char not in TAG_CHARS)


def get_invalid_path_chars(path):
    """Return any invalid characters in the directory path.

    :param path: the directory path
    :returns: set of all invalid characters
    """
    return set(char for char in path if char not in PATH_CHARS)



from __future__ import unicode_literals

import os
import sys


def finish(message):
    """Write the message to stdout and exit with 0.

    :param message: the message to write
    """
    sys.stdout.write(message + '\n')
    sys.exit(0)


def warn(message):
    """Write the warning message to stderr.

    :param message: the warning message to write
    """
    sys.stderr.write(message + '\n')


def abort(message, exit_code=1):
    """Write the message to stderr and exit with the given exit code.

    :param message: the error message to write
    :param exit_code: the exit code
    """
    sys.stderr.write(message + '\n')
    sys.exit(exit_code)


def expand(path):
    """Fully expand the given file path.

    :param path: the file path to expand
    :return: the expanded path
    """
    return os.path.abspath(os.path.expanduser(path))


def close_stdio():
    """Close stdout and stderr."""
    try:
        sys.stdout.close()
    except IOError:
        pass
    try:
        sys.stderr.close()
    except IOError:
        pass


def rm_files(*filenames):
    """Remove the files from disk while ignoring ENOENT.

    :param filenames: the files to remove
    """
    for filename in filenames:
        try:
            os.remove(filename)
        except OSError as remove_error:
            if remove_error.errno != 2:  # ENOENT
                raise remove_error



from __future__ import unicode_literals, print_function

import os
import sys
import atexit
import signal
import tempfile
import collections
import subprocess as sp

from dtags import style
from dtags.config import load_mapping
from dtags.utils import close_stdio, abort, finish, expand
from dtags.version import VERSION

temp_file = None
process = None
processes = []

USAGE = """Usage:
  e [-p] <targets> <command> [<arg>...]
  e --help
  e --version
"""
DESCRIPTION = """
Arguments:
  targets     Comma separated directory tags or paths
  command     The command to execute
  arg         The command arguments

Options:
  -p          Execute the commands in parallel
  --help      Display the help menu
  --version   Display the version

Execute commands in one or more directories.
Multiple targets can be given by separating
them with commas and without whitespaces."""


def _send_sigterm(child_process):
    """Send SIGTERM to the specified child process."""
    try:
        os.killpg(os.getpgid(child_process.pid), signal.SIGKILL)
    except OSError as kill_error:
        # If the process is already dead, nothing needs to be done
        if kill_error.errno != 3:  # ESRCH
            raise kill_error


def _cleanup_resources():
    """Kill all child processes and close any open files."""
    global temp_file, process, processes

    if temp_file is not None:
        temp_file.close()
    if process is not None:
        _send_sigterm(process)
    for _, _, child_process, open_file in processes:
        _send_sigterm(child_process)
        if open_file is not None:
            open_file.close()
    close_stdio()


def _handle_signal(signum, *_):
    """Send SIGTERM to all child processes."""
    _cleanup_resources()
    sys.exit(127 + signum)


def main():
    atexit.register(_cleanup_resources)
    signal.signal(signal.SIGPIPE, signal.SIG_DFL)
    mapping, excluded = load_mapping()
    if excluded:
        print(
            'Found invalid entries in the mapping:\n' +
            excluded + '\nRun ' + style.cmd('dtags clean') +
            ' to remove them' + '\n'
        )
    # Parse the optional arguments
    parallel = False
    args = sys.argv[1:]
    if not args:
        finish(USAGE + DESCRIPTION)
    head, tail = args[0], args[1:]
    if head == '--help':
        finish(USAGE + DESCRIPTION)
    elif head == '--version':
        finish('Version ' + VERSION)
    elif head == '-p' and not tail:
        abort(USAGE + 'Missing argument: ' + style.bad('<targets>'))
    elif head == '-p' and tail:
        parallel = True
        head, tail = tail[0], tail[1:]
    elif head.startswith('-'):
        abort(USAGE + 'Invalid argument: ' + style.bad(head))

    # Parse the positional arguments
    if not tail:
        abort(USAGE + 'Missing argument: ' + style.bad('<command>'))
    directories = collections.defaultdict(set)
    for target in head.split(','):
        if not target:
            continue
        elif target in mapping:
            for directory in sorted(mapping[target]):
                directories[directory].add(target)
        else:
            path = expand(target)
            if os.path.isdir(path):
                directories[path].add(None)
            else:
                abort(USAGE + 'Invalid target: ' + style.bad(target))
    command = sp.list2cmdline(tail)

    # Check which shell is in use (e.g. zsh, bash, fish)
    shell = os.environ.get('SHELL')
    if shell is None:
        abort('Undefined environment variable: ' + style.bad('SHELL'))

    # Execute the command in the targeted directories
    msg_head = style.msg('Executing command ') + style.cmd(command)
    if parallel:
        global temp_file, process, processes

        # Add signal handlers to terminate child processes gracefully
        signal.signal(signal.SIGINT, _handle_signal)
        signal.signal(signal.SIGABRT, _handle_signal)
        signal.signal(signal.SIGTERM, _handle_signal)

        # Execute in parallel and pipe the output to temporary files
        sys.stdout.write(msg_head + style.msg(' in parallel...\n\n'))
        for directory in sorted(directories):
            tags = directories[directory]
            temp_file = tempfile.TemporaryFile(mode='w+t')
            process = sp.Popen(
                [shell, '-i', '-c', command],
                cwd=directory,
                stdout=temp_file,
                stderr=temp_file,
                preexec_fn=os.setsid
            )
            processes.append((directory, tags, process, temp_file))

        # Read from the temporary files back to stdout line by line
        for directory, tags, process, temp_file in processes:
            status = process.wait()
            tags = [style.tag(tag) for tag in tags if tag]
            sys.stdout.write('{} {}{}\n'.format(
                style.msg('in'), style.path(directory),
                ((' ' + ' '.join(tags)) if tags else '') + style.msg(':')
            ))
            temp_file.seek(0)
            for line in temp_file:
                sys.stdout.write(line)
            temp_file.close()
            sys.stdout.write(style.msg('Exit status: {}\n\n'.format(status)))
    else:
        # Generate the command string and execute all in one subprocess call
        commands = []
        status_printf = 'printf "{}\n\n"'.format(style.msg(
            'Exit status: ' + ('$status' if '/fish' in shell else '$?')
        ))
        for directory in sorted(directories):
            tags = [style.tag(tag) for tag in directories[directory] if tag]
            commands.append('printf "{} {}{}\n"; cd "{}"; {};{}'.format(
                style.msg('in'), style.path(directory),
                ((' ' + ' '.join(tags)) if tags else '') + style.msg(':'),
                directory, command, status_printf
            ))
        sys.stdout.write(msg_head + style.msg(' in sequence...\n\n'))
        sys.stdout.flush()  # flush for printing chronologically
        sp.call([shell, '-i', '-c', ';'.join(commands)], stderr=sp.STDOUT)
